#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import os
import signal
from typing import Optional

from airflow.configuration import conf
from airflow.events.scheduler_events import TaskStateChangedEvent
from airflow.exceptions import AirflowException
from airflow.jobs.base_job import BaseJob
from airflow.models.taskinstance import TaskInstance
from airflow.stats import Stats
from airflow.task.task_runner import get_task_runner
from airflow.utils import timezone
from airflow.utils.net import get_hostname
from airflow.utils.session import provide_session
from airflow.utils.state import State
from notification_service.client import NotificationClient


class LocalTaskJob(BaseJob):
    """LocalTaskJob runs a single task instance."""

    __mapper_args__ = {'polymorphic_identity': 'LocalTaskJob'}

    def __init__(
        self,
        task_instance: TaskInstance,
        ignore_all_deps: bool = False,
        ignore_depends_on_past: bool = False,
        ignore_task_deps: bool = False,
        ignore_ti_state: bool = False,
        mark_success: bool = False,
        pickle_id: Optional[str] = None,
        pool: Optional[str] = None,
        server_uri: str = None,
        *args,
        **kwargs,
    ):
        self.task_instance = task_instance
        self.dag_id = task_instance.dag_id
        self.ignore_all_deps = ignore_all_deps
        self.ignore_depends_on_past = ignore_depends_on_past
        self.ignore_task_deps = ignore_task_deps
        self.ignore_ti_state = ignore_ti_state
        self.pool = pool
        self.pickle_id = pickle_id
        self.mark_success = mark_success
        self.task_runner = None
        self.server_uri = server_uri

        # terminating state is used so that a job don't try to
        # terminate multiple times
        self.terminating = False

        super().__init__(*args, **kwargs)

    def _execute(self):
        self.task_runner = get_task_runner(self)

        # pylint: disable=unused-argument
        def signal_handler(signum, frame):
            """Setting kill signal handler"""
            self.log.error("Received SIGTERM. Terminating subprocesses")
            self.on_kill()
            raise AirflowException("LocalTaskJob received SIGTERM signal")

        # pylint: enable=unused-argument
        signal.signal(signal.SIGTERM, signal_handler)

        if not self.task_instance.check_and_change_state_before_execution(
            mark_success=self.mark_success,
            ignore_all_deps=self.ignore_all_deps,
            ignore_depends_on_past=self.ignore_depends_on_past,
            ignore_task_deps=self.ignore_task_deps,
            ignore_ti_state=self.ignore_ti_state,
            job_id=self.id,
            pool=self.pool,
        ):
            self.log.info("Task is not able to be run")
            return

        try:
            if self.server_uri is not None:
                try:
                    self._send_task_status_change_event()
                except Exception as e:
                    self.log.warning("failed to send event to {}".format(self.server_uri), exc_info=e)
            self.task_runner.start()

            heartbeat_time_limit = conf.getint('scheduler', 'scheduler_zombie_task_threshold')

            while True:
                # Monitor the task to see if it's done. Wait in a syscall
                # (`os.wait`) for as long as possible so we notice the
                # subprocess finishing as quick as we can
                max_wait_time = max(
                    0,  # Make sure this value is never negative,
                    min(
                        (
                            heartbeat_time_limit
                            - (timezone.utcnow() - self.latest_heartbeat).total_seconds() * 0.75
                        ),
                        self.heartrate,
                    ),
                )

                return_code = self.task_runner.return_code(timeout=max_wait_time)
                if return_code is not None:
                    self.log.info("Task exited with return code %s", return_code)
                    return

                self.heartbeat()

                # If it's been too long since we've heartbeat, then it's possible that
                # the scheduler rescheduled this task, so kill launched processes.
                # This can only really happen if the worker can't read the DB for a long time
                time_since_last_heartbeat = (timezone.utcnow() - self.latest_heartbeat).total_seconds()
                if time_since_last_heartbeat > heartbeat_time_limit:
                    Stats.incr('local_task_job_prolonged_heartbeat_failure', 1, 1)
                    self.log.error("Heartbeat time limit exceeded!")
                    raise AirflowException(
                        "Time since last heartbeat({:.2f}s) "
                        "exceeded limit ({}s).".format(time_since_last_heartbeat, heartbeat_time_limit)
                    )
        finally:
            self.on_kill()

    def _send_task_status_change_event(self):
        task_status_changed_event = TaskStateChangedEvent(
            self.task_instance.task_id,
            self.task_instance.dag_id,
            self.task_instance.execution_date,
            self.task_instance.state
        )
        event = task_status_changed_event.to_event()
        client = NotificationClient(self.server_uri, default_namespace=event.namespace, sender=event.sender)
        self.log.info("LocalTaskJob sending event: {}".format(event))
        client.send_event(event)

    def on_kill(self):
        self.task_runner.terminate()
        self.task_runner.on_finish()

    @provide_session
    def heartbeat_callback(self, session=None):
        """Self destruct task if state has been moved away from running externally"""
        if self.terminating:
            # ensure termination if processes are created later
            self.task_runner.terminate()
            return

        self.task_instance.refresh_from_db()
        ti = self.task_instance

        if ti.state == State.RUNNING:
            fqdn = get_hostname()
            same_hostname = fqdn == ti.hostname
            if not same_hostname:
                self.log.warning(
                    "The recorded hostname %s " "does not match this instance's hostname " "%s",
                    ti.hostname,
                    fqdn,
                )
                raise AirflowException("Hostname of job runner does not match")

            current_pid = os.getpid()
            same_process = ti.pid == current_pid
            if not same_process:
                self.log.warning("Recorded pid %s does not match " "the current pid %s", ti.pid, current_pid)
                raise AirflowException("PID of job runner does not match")
        elif ti.state == State.KILLING:
            self.log.warning(
                "This instance is being killed %s", ti)
        elif self.task_runner.return_code() is None and hasattr(self.task_runner, 'process'):
            self.log.warning(
                "State of this instance has been externally set to %s. " "Terminating instance.", ti.state
            )
            if ti.state == State.FAILED and ti.task.on_failure_callback:
                context = ti.get_template_context()
                ti.task.on_failure_callback(context)
            if ti.state == State.SUCCESS and ti.task.on_success_callback:
                context = ti.get_template_context()
                ti.task.on_success_callback(context)
            self.task_runner.terminate()
            self.terminating = True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Text, List

from ai_flow.meta.dataset_meta import DatasetMeta

from ai_flow.graph.node import Node
from ai_flow.workflow.job_config import JobConfig


class Job(Node):
    """
    Job is a description of an executable unit,
    which contains information about the configuration and data required to run.
    """
    def __init__(self,
                 job_config: JobConfig) -> None:
        """
        :param job_config: Job configuration information(ai_flow.workflow.job_config.JobConfig).
        """
        super().__init__()
        self.job_config = job_config
        self.input_dataset_list: List[DatasetMeta] = []  # the job read dataset information
        self.output_dataset_list: List[DatasetMeta] = []  # the job write dataset information

    @property
    def job_name(self):
        """The name of the job."""
        return self.job_config.job_name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import Sequence, Union


class TaskMixin:
    """
    Mixing implementing common chain methods like >> and <<.

    In the following functions we use:
    Task = Union[BaseOperator, XComArg]
    No type annotations due to cyclic imports.
    """

    @property
    def roots(self):
        """Should return list of root operator List[BaseOperator]"""
        raise NotImplementedError()

    @property
    def leaves(self):
        """Should return list of leaf operator List[BaseOperator]"""
        raise NotImplementedError()

    @abstractmethod
    def set_upstream(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Set a task or a task list to be directly upstream from the current task."""
        raise NotImplementedError()

    @abstractmethod
    def set_downstream(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Set a task or a task list to be directly downstream from the current task."""
        raise NotImplementedError()

    def update_relative(self, other: "TaskMixin", upstream=True) -> None:
        """
        Update relationship information about another TaskMixin. Default is no-op.
        Override if necessary.
        """

    def __lshift__(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Implements Task << Task"""
        self.set_upstream(other)
        return other

    def __rshift__(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Implements Task >> Task"""
        self.set_downstream(other)
        return other

    def __rrshift__(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Called for Task >> [Task] because list don't have __rshift__ operators."""
        self.__lshift__(other)
        return self

    def __rlshift__(self, other: Union["TaskMixin", Sequence["TaskMixin"]]):
        """Called for Task << [Task] because list don't have __lshift__ operators."""
        self.__rshift__(other)
        return self



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Text

from ai_flow.util.json_utils import Jsonable


class Edge(Jsonable):
    """
    Edges are part of the graph(ai_flow.graph.graph.Graph),
    and there are edges connected between nodes(ai_flow.graph.node.Node).
    """

    def __init__(self,
                 source: Text,
                 destination: Text,
                 ) -> None:
        """
        :param source: source identifies the starting node's node_id of the directed edge.
        :param destination: destination identifies the ending node's node_id of the directed edge.
        """
        super().__init__()
        if source is None or destination is None:
            raise Exception('destination node id or source node id can not be None!')
        self.source = source
        self.destination = destination

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Edge):
            return self.destination == o.destination and self.source == o.source
        else:
            return False

    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
import os
from ai_flow.util.path_util import get_file_dir
from ai_flow.project.project_config import ProjectConfig


class TestProjectConfig(unittest.TestCase):

    def test_load_project_config(self):
        project_path = get_file_dir(__file__)
        project_config = ProjectConfig()
        project_config.load_from_file(os.path.join(project_path, 'project.yaml'))
        self.assertEqual(project_config.get_server_uri(), "localhost:50051")
        self.assertIsNone(project_config.get('ai_flow config', None))
        self.assertEqual(project_config['ai_flow_home'], '/opt/ai_flow')
        self.assertEqual(project_config['ai_flow_job_master.host'], 'localhost')
        self.assertEqual(project_config['ai_flow_job_master.port'], 8081)
        self.assertEqual(project_config['ai_flow_conf'], 'taskmanager.slot=2')


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""Utilities for creating a virtual environment"""
import os
from typing import List, Optional

import jinja2

from airflow.utils.process_utils import execute_in_subprocess


def _generate_virtualenv_cmd(tmp_dir: str, python_bin: str, system_site_packages: bool) -> List[str]:
    cmd = ['virtualenv', tmp_dir]
    if system_site_packages:
        cmd.append('--system-site-packages')
    if python_bin is not None:
        cmd.append(f'--python={python_bin}')
    return cmd


def _generate_pip_install_cmd(tmp_dir: str, requirements: List[str]) -> Optional[List[str]]:
    if not requirements:
        return None
    # direct path alleviates need to activate
    cmd = [f'{tmp_dir}/bin/pip', 'install']
    return cmd + requirements


def prepare_virtualenv(
    venv_directory: str, python_bin: str, system_site_packages: bool, requirements: List[str]
) -> str:
    """
    Creates a virtual environment and installs the additional python packages

    :param venv_directory: The path for directory where the environment will be created
    :type venv_directory: str
    :param python_bin: Path for python binary
    :type python_bin: str
    :param system_site_packages: Whether to include system_site_packages in your virtualenv.
        See virtualenv documentation for more information.
    :type system_site_packages: bool
    :param requirements: List of additional python packages
    :type requirements: List[str]
    :return: Path to a binary file with Python in a virtual environment.
    :rtype: str
    """
    virtualenv_cmd = _generate_virtualenv_cmd(venv_directory, python_bin, system_site_packages)
    execute_in_subprocess(virtualenv_cmd)
    pip_cmd = _generate_pip_install_cmd(venv_directory, requirements)
    if pip_cmd:
        execute_in_subprocess(pip_cmd)

    return f'{venv_directory}/bin/python'


def write_python_script(jinja_context: dict, filename: str):
    """
    Renders the python script to a file to execute in the virtual environment.

    :param jinja_context: The jinja context variables to unpack and replace with its placeholders in the
        template file.
    :type jinja_context: dict
    :param filename: The name of the file to dump the rendered script to.
    :type filename: str
    """
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.dirname(__file__))
    template_env = jinja2.Environment(loader=template_loader, undefined=jinja2.StrictUndefined)
    template = template_env.get_template('python_virtualenv_script.jinja2')
    template.stream(**jinja_context).dump(filename)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.model_center.entity._model_repo_entity import _ModelRepoEntity


class RegisteredModelParam(_ModelRepoEntity):
    """
    AIFlow entity for Registered Model Parameter.
    """

    def __init__(self, model_name, model_desc):
        self._model_name = model_name
        self._model_desc = model_desc

    @property
    def model_name(self):
        """String. Unique name for this registered model within Model Registry."""
        return self._model_name

    @property
    def model_desc(self):
        """String. Model desc for this registered model within Model Registry."""
        return self._model_desc

    # proto mappers
    @classmethod
    def from_proto(cls, proto):
        registered_model = proto.registered_model
        return cls(registered_model.model_name.value if registered_model.HasField("model_name") else None,
                   registered_model.model_desc.value if registered_model.HasField("model_desc") else None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import date, datetime

import numpy as np
from flask.json import JSONEncoder

try:
    from kubernetes.client import models as k8s
except ImportError:
    k8s = None

# Dates and JSON encoding/decoding


class AirflowJsonEncoder(JSONEncoder):
    """Custom Airflow json encoder implementation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self._default

    @staticmethod
    def _default(obj):
        """Convert dates and numpy objects in a json serializable format."""
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(obj, date):
            return obj.strftime('%Y-%m-%d')
        elif isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(
            obj, (np.float_, np.float16, np.float32, np.float64, np.complex_, np.complex64, np.complex128)
        ):
            return float(obj)
        elif k8s is not None and isinstance(obj, k8s.V1Pod):
            from airflow.kubernetes.pod_generator import PodGenerator

            return PodGenerator.serialize_pod(obj)

        raise TypeError(f"Object of type '{obj.__class__.__name__}' is not JSON serializable")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import inspect
import os
from copy import copy
from functools import wraps
from typing import Any, Callable, Dict, TypeVar, cast

from airflow.exceptions import AirflowException

signature = inspect.signature

T = TypeVar('T', bound=Callable)  # pylint: disable=invalid-name


def apply_defaults(func: T) -> T:
    """
    Function decorator that Looks for an argument named "default_args", and
    fills the unspecified arguments from it.

    Since python2.* isn't clear about which arguments are missing when
    calling a function, and that this can be quite confusing with multi-level
    inheritance and argument defaults, this decorator also alerts with
    specific information about the missing arguments.
    """
    # Cache inspect.signature for the wrapper closure to avoid calling it
    # at every decorated invocation. This is separate sig_cache created
    # per decoration, i.e. each function decorated using apply_defaults will
    # have a different sig_cache.
    sig_cache = signature(func)
    non_optional_args = {
        name
        for (name, param) in sig_cache.parameters.items()
        if param.default == param.empty
        and param.name != 'self'
        and param.kind not in (param.VAR_POSITIONAL, param.VAR_KEYWORD)
    }

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        from airflow.models.dag import DagContext

        if len(args) > 1:
            raise AirflowException("Use keyword arguments when initializing operators")
        dag_args: Dict[str, Any] = {}
        dag_params: Dict[str, Any] = {}

        dag = kwargs.get('dag') or DagContext.get_current_dag()
        if dag:
            dag_args = copy(dag.default_args) or {}
            dag_params = copy(dag.params) or {}

        params = kwargs.get('params', {}) or {}
        dag_params.update(params)

        default_args = {}
        if 'default_args' in kwargs:
            default_args = kwargs['default_args']
            if 'params' in default_args:
                dag_params.update(default_args['params'])
                del default_args['params']

        dag_args.update(default_args)
        default_args = dag_args

        for arg in sig_cache.parameters:
            if arg not in kwargs and arg in default_args:
                kwargs[arg] = default_args[arg]

        missing_args = list(non_optional_args - set(kwargs))
        if missing_args:
            msg = f"Argument {missing_args} is required"
            raise AirflowException(msg)

        kwargs['params'] = dag_params

        result = func(*args, **kwargs)
        return result

    return cast(T, wrapper)


if 'BUILDING_AIRFLOW_DOCS' in os.environ:
    # flake8: noqa: F811
    # Monkey patch hook to get good function headers while building docs
    apply_defaults = lambda x: x



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.meta.workflow_meta import WorkflowMeta
from typing import Optional, Text, List

import grpc

from ai_flow.common.status import Status
from ai_flow.meta.artifact_meta import ArtifactMeta
from ai_flow.meta.dataset_meta import DatasetMeta, Properties, DataType
from ai_flow.meta.model_meta import ModelMeta, ModelVersionMeta
from ai_flow.meta.model_relation_meta import ModelRelationMeta, ModelVersionRelationMeta
from ai_flow.meta.project_meta import ProjectMeta
from ai_flow.metadata_store.utils.MetaToProto import MetaToProto
from ai_flow.protobuf import metadata_service_pb2_grpc, metadata_service_pb2
from ai_flow.protobuf.message_pb2 import DatasetProto, SchemaProto, ModelRelationProto, ModelProto, \
    ModelVersionRelationProto, ModelVersionProto, ProjectProto, \
    ArtifactProto, ModelVersionStage, WorkflowMetaProto
from ai_flow.protobuf.metadata_service_pb2 import ModelNameRequest
from ai_flow.endpoint.server import stringValue, int64Value
from ai_flow.endpoint.client.base_client import BaseClient
from ai_flow.endpoint.server.util import _unwrap_dataset_response, \
    transform_dataset_type_list_to_proto, _unwrap_dataset_list_response, _unwrap_delete_response, \
    _unwrap_model_relation_response, _unwrap_model_relation_list_response, _unwrap_model_response, \
    _unwrap_model_version_relation_response, _unwrap_model_version_relation_list_response, \
    _unwrap_model_version_response, \
    _unwrap_project_response, _unwrap_project_list_response, \
    _unwrap_artifact_response, _unwrap_artifact_list_response, _unwrap_workflow_response, _unwrap_workflow_list_response


class MetadataClient(BaseClient):
    def __init__(self, server_uri):
        super(MetadataClient, self).__init__(server_uri)
        channel = grpc.insecure_channel(server_uri)
        self.metadata_store_stub = metadata_service_pb2_grpc.MetadataServiceStub(channel)

    '''dataset api'''

    def get_dataset_by_id(self, dataset_id) -> Optional[DatasetMeta]:
        """
        get a specific dataset in metadata store by dataset id.

        :param dataset_id: the dataset id
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,
                 Otherwise, returns None if the dataset does not exist.
        """
        request = metadata_service_pb2.IdRequest(id=dataset_id)
        response = self.metadata_store_stub.getDatasetById(request)
        return _unwrap_dataset_response(response)

    def get_dataset_by_name(self, dataset_name) -> Optional[DatasetMeta]:
        """
        get a specific dataset in metadata store by dataset name.

        :param dataset_name: the dataset name
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,,
                 Otherwise, returns None if the dataset does not exist.
        """
        request = metadata_service_pb2.NameRequest(name=dataset_name)
        response = self.metadata_store_stub.getDatasetByName(request)
        return _unwrap_dataset_response(response)

    def register_dataset(self, name: Text, data_format: Text = None, description: Text = None,
                         uri: Text = None, properties: Properties = None, name_list: List[Text] = None,
                         type_list: List[DataType] = None) -> DatasetMeta:
        """
        register an dataset in metadata store.

        :param name: the name of the dataset
        :param data_format: the data format of the dataset
        :param description: the description of the dataset
        :param uri: the uri of the dataset
        :param properties: the properties of the dataset
        :param name_list: the name list of dataset's schema
        :param type_list: the type list corresponded to the name list of dataset's schema
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        request = metadata_service_pb2.RegisterDatasetRequest(
            dataset=DatasetProto(name=name,  data_format=stringValue(data_format),
                                 description=stringValue(description), uri=stringValue(uri),
                                 properties=properties,
                                 schema=SchemaProto(name_list=name_list, type_list=transform_dataset_type_list_to_proto(
                                     type_list)),
                                 catalog_name=stringValue(None), catalog_type=stringValue(None),
                                 catalog_database=stringValue(None), catalog_connection_uri=stringValue(None),
                                 catalog_table=stringValue(None)))
        response = self.metadata_store_stub.registerDataset(request)
        return _unwrap_dataset_response(response)

    def register_dataset_with_catalog(self, name: Text,
                                      catalog_name: Text, catalog_type: Text,
                                      catalog_connection_uri: Text,
                                      catalog_table: Text, catalog_database: Text = None) -> DatasetMeta:
        """
        register dataset with catalog in metadata store.

        :param name: the name of the dataset
        :param catalog_name: the name of the dataset catalog
        :param catalog_type: the type of the dataset catalog
        :param catalog_connection_uri: the connection uri of the dataset catalog
        :param catalog_table: the table of the dataset catalog
        :param catalog_database: the database of the dataset catalog
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        request = metadata_service_pb2.RegisterDatasetRequest(
            dataset=DatasetProto(name=name,
                                 data_format=stringValue(None), description=stringValue(None),
                                 uri=stringValue(None),
                                 properties=None,
                                 schema=SchemaProto(name_list=None, type_list=transform_dataset_type_list_to_proto(
                                     None)),
                                 catalog_name=stringValue(catalog_name), catalog_type=stringValue(catalog_type),
                                 catalog_database=stringValue(catalog_database),
                                 catalog_connection_uri=stringValue(catalog_connection_uri),
                                 catalog_table=stringValue(catalog_table)))
        response = self.metadata_store_stub.registerDatasetWithCatalog(request)
        return _unwrap_dataset_response(response)

    def register_datasets(self, datasets: List[DatasetMeta]) -> List[DatasetMeta]:
        """
        register multiple datasets in metadata store.

        :param datasets: A list of datasets
        :return: List of :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` objects.
        """
        request = metadata_service_pb2.RegisterDatasetsRequest(
            datasets=MetaToProto.dataset_meta_list_to_proto(datasets))
        response = self.metadata_store_stub.registerDatasets(request)
        return _unwrap_dataset_list_response(response)

    def update_dataset(self, dataset_name: Text, data_format: Text = None,
                       description: Text = None, uri: Text = None,
                       properties: Properties = None, name_list: List[Text] = None,
                       type_list: List[DataType] = None, catalog_name: Text = None,
                       catalog_type: Text = None, catalog_database: Text = None,
                       catalog_connection_uri: Text = None,
                       catalog_table: Text = None) -> Optional[DatasetMeta]:
        """
        update dataset in metadata store.

        :param dataset_name: the name of the dataset
        :param data_format: the data format of the dataset
        :param description: the description of the dataset
        :param uri: the uri of the dataset
        :param properties: the properties of the dataset
        :param name_list: the name list of dataset's schema
        :param type_list: the type list corresponded to the name list of dataset's schema
        :param catalog_name: the name of the dataset catalog
        :param catalog_type: the type of the dataset catalog
        :param catalog_database: :param catalog_database: the database of the dataset catalog
        :param catalog_connection_uri: the connection uri of the dataset catalog
        :param catalog_table: the table of the dataset catalog
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if update successfully.
        """
        request = metadata_service_pb2.UpdateDatasetRequest(name=dataset_name,
                                                            data_format=stringValue(data_format),
                                                            description=stringValue(description),
                                                            uri=stringValue(uri),
                                                            properties=properties,
                                                            name_list=name_list,
                                                            type_list=transform_dataset_type_list_to_proto(
                                                                type_list),
                                                            catalog_name=stringValue(catalog_name),
                                                            catalog_type=stringValue(catalog_type),
                                                            catalog_database=stringValue(catalog_database),
                                                            catalog_connection_uri=stringValue(catalog_connection_uri),
                                                            catalog_table=stringValue(catalog_table))
        response = self.metadata_store_stub.updateDataset(request)
        return _unwrap_dataset_response(response)

    def list_datasets(self, page_size, offset) -> Optional[List[DatasetMeta]]:
        """
        List registered datasets in metadata store.

        :param page_size: the limitation of the listed datasets.
        :param offset: the offset of listed datasets.
        :return: List of :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` objects,
                 return None if no datasets to be listed.
        """
        request = metadata_service_pb2.ListRequest(page_size=page_size, offset=offset)
        response = self.metadata_store_stub.listDatasets(request)
        return _unwrap_dataset_list_response(response)

    def delete_dataset_by_name(self, dataset_name) -> Status:
        """
        Delete the registered dataset by dataset name .

        :param dataset_name: the dataset name
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        request = metadata_service_pb2.NameRequest(name=dataset_name)
        response = self.metadata_store_stub.deleteDatasetByName(request)
        return _unwrap_delete_response(response)

    def delete_dataset_by_id(self, dataset_id):
        """
        Delete the registered dataset by dataset id .

        :param dataset_id: the dataset id
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        request = metadata_service_pb2.IdRequest(id=dataset_id)
        response = self.metadata_store_stub.deleteDatasetByName(request)
        return _unwrap_delete_response(response)

    '''model relation api'''

    def get_model_relation_by_id(self, model_id) -> Optional[ModelRelationMeta]:
        """
        get a specific model relation in metadata store by model id.

        :param model_id: the model id
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        request = metadata_service_pb2.IdRequest(id=model_id)
        response = self.metadata_store_stub.getModelRelationById(request)
        return _unwrap_model_relation_response(response)

    def get_model_relation_by_name(self, model_name) -> Optional[ModelRelationMeta]:
        """
        get a specific model relation in metadata store by model name.

        :param model_name: the model name
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        request = metadata_service_pb2.NameRequest(name=model_name)
        response = self.metadata_store_stub.getModelRelationByName(request)
        return _unwrap_model_relation_response(response)

    def register_model_relation(self, name, project_id) -> ModelRelationMeta:
        """
        register a model relation in metadata store

        :param name: the name of the model
        :param project_id: the project id which the model corresponded to.
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object.
        """
        request = metadata_service_pb2.RegisterModelRelationRequest(
            model_relation=ModelRelationProto(name=name, project_id=int64Value(project_id)))
        response = self.metadata_store_stub.registerModelRelation(request)
        return _unwrap_model_relation_response(response)

    def list_model_relation(self, page_size, offset) -> List[ModelRelationMeta]:
        """
        List registered model relations in metadata store.

        :param page_size: the limitation of the listed model relations.
        :param offset: the offset of listed model relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model relations to be listed.
        """
        request = metadata_service_pb2.ListRequest(page_size=page_size, offset=offset)
        response = self.metadata_store_stub.listModelRelation(request)
        return _unwrap_model_relation_list_response(response)

    def delete_model_relation_by_id(self, model_id) -> Status:
        """
        Delete the registered model by model id .

        :param model_id: the model id
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        request = metadata_service_pb2.IdRequest(id=model_id)
        response = self.metadata_store_stub.deleteModelRelationById(request)
        return _unwrap_delete_response(response)

    def delete_model_relation_by_name(self, model_name) -> Status:
        """
        Delete the registered model by model name .

        :param model_name: the model name
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        request = metadata_service_pb2.NameRequest(name=model_name)
        response = self.metadata_store_stub.deleteModelRelationByName(request)
        return _unwrap_delete_response(response)

    '''model api'''

    def get_model_by_id(self, model_id) -> Optional[ModelMeta]:
        """
        get a specific model in metadata store by model id.

        :param model_id: Id of registered model
        :return: A single :py:class:`ai_flow.meta.model_meta.ModelMeta` object if the model relation exists,
        Otherwise, returns None if the model relation does not exist.
        """
        request = metadata_service_pb2.IdRequest(id=model_id)
        response = self.metadata_store_stub.getModelById(request)
        return _unwrap_model_response(response)

    def get_model_by_name(self, model_name) -> Optional[ModelMeta]:
        """
        get a specific model in metadata store by model name.

        :param model_name: Name of registered model
        :return: A single :py:class:`ai_flow.meta.model_meta.ModelMeta` object if the model relation exists,
        Otherwise, returns None if the model relation does not exist.
        """
        request = metadata_service_pb2.NameRequest(name=model_name)
        response = self.metadata_store_stub.getModelByName(request)
        return _unwrap_model_response(response)

    def register_model(self, model_name, project_id, model_desc=None) -> ModelMeta:
        """
        register a model in metadata store

        :param model_name: Name of registered model
        :param project_id: Project id which registered model corresponded to.
        :param model_desc: Description of registered model
        :return: A single :py:class:`ai_flow.meta.model_meta.ModelMeta` object.
        """
        model_request = ModelProto(name=model_name,
                                   model_desc=stringValue(model_desc),
                                   project_id=int64Value(project_id))
        request = metadata_service_pb2.RegisterModelRequest(model=model_request)
        response = self.metadata_store_stub.registerModel(request)
        return _unwrap_model_response(response)

    def delete_model_by_id(self, model_id) -> Status:
        """
        delete registered model by model id.

        :param model_id: Id of registered model
        :return: Status.OK if registered model is successfully deleted,
                 Status.ERROR if registered model does not exist otherwise.
        """
        request = metadata_service_pb2.IdRequest(id=model_id)
        response = self.metadata_store_stub.deleteModelById(request)
        return _unwrap_delete_response(response)

    def delete_model_by_name(self, model_name) -> Status:
        """
        delete registered model by model name.

        :param model_name: Name of registered model
        :return: Status.OK if registered model is successfully deleted,
                 Status.ERROR if registered model does not exist otherwise.
        """
        request = metadata_service_pb2.NameRequest(name=model_name)
        response = self.metadata_store_stub.deleteModelByName(request)
        return _unwrap_delete_response(response)

    '''model version relation api'''

    def get_model_version_relation_by_version(self, version, model_id) -> Optional[ModelVersionRelationMeta]:
        """
        get a specific model version relation in metadata store by the model version name.

        :param version: the model version name
        :param model_id: the model id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object
                 if the model version exists, Otherwise, returns None if the model version does not exist.
        """
        request = metadata_service_pb2.ModelVersionNameRequest(name=version, model_id=model_id)
        response = self.metadata_store_stub.getModelVersionRelationByVersion(request)
        return _unwrap_model_version_relation_response(response)

    def register_model_version_relation(self, version, model_id,
                                        project_snapshot_id=None) -> ModelVersionRelationMeta:
        """
        register a model version relation in metadata store.

        :param version: the specific model version
        :param model_id: the model id corresponded to the model version
        :param project_snapshot_id: the project snapshot id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object.
        """
        model_version = ModelVersionRelationProto(version=stringValue(version), model_id=int64Value(model_id),
                                                  project_snapshot_id=int64Value(project_snapshot_id))
        request = metadata_service_pb2.RegisterModelVersionRelationRequest(model_version_relation=model_version)
        response = self.metadata_store_stub.registerModelVersionRelation(request)
        return _unwrap_model_version_relation_response(response)

    def list_model_version_relation(self, model_id, page_size, offset) -> List[ModelVersionRelationMeta]:
        """
        List registered model version relations in metadata store.

        :param model_id: the model id corresponded to the model version
        :param page_size: the limitation of the listed model version relations.
        :param offset: the offset of listed model version relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model version relations to be listed.
        """
        request = metadata_service_pb2.ListModelVersionRelationRequest(model_id=model_id, page_size=page_size,
                                                                       offset=offset)
        response = self.metadata_store_stub.listModelVersionRelation(request)
        return _unwrap_model_version_relation_list_response(response)

    def delete_model_version_relation_by_version(self, version, model_id) -> Status:
        """
        Delete the registered model version by model version name .

        :param version: the model version name
        :param model_id: the model id corresponded to the model version
        :return: Status.OK if the model version is successfully deleted,
                 Status.ERROR if the model version does not exist otherwise.
        """
        request = metadata_service_pb2.ModelVersionNameRequest(name=version, model_id=model_id)
        response = self.metadata_store_stub.deleteModelVersionRelationByVersion(request)
        return _unwrap_delete_response(response)

    '''model version api'''

    def get_model_version_by_version(self, version, model_id) -> Optional[ModelVersionMeta]:
        """
        Get a specific model version in metadata store by model version name.

        :param version: User-defined version of registered model
        :param model_id: The model id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_meta.ModelVersionMeta` object if the model version exists,
        Otherwise, returns None if the model version does not exist.
        """
        request = metadata_service_pb2.ModelVersionNameRequest(name=version, model_id=model_id)
        response = self.metadata_store_stub.getModelVersionByVersion(request)
        return _unwrap_model_version_response(response)

    def register_model_version(self, model, model_path, project_snapshot_id=None,
                               model_type=None, version_desc=None,
                               current_stage=ModelVersionStage.GENERATED) -> ModelVersionMeta:
        """
        register a model version in metadata store.

        :param model: Model id or model meta of registered model corresponded to model version
        :param model_path: Source path where the AIFlow model is stored.
        :param project_snapshot_id: Id of project snapshot corresponded to model version
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) Description of registered model version.
        :param current_stage: (Optional) Stage of registered model version
        :return: A single :py:class:`ai_flow.meta.model_meta.ModelVersionMeta` object.
        """

        if isinstance(model, int):
            model_id = model
        elif isinstance(model, ModelMeta):
            model_id = model.uuid
        else:
            raise Exception("can not recognize model {}".format(model))
        model_version = ModelVersionProto(version=None,
                                          model_id=int64Value(model_id),
                                          project_snapshot_id=int64Value(project_snapshot_id),
                                          model_path=stringValue(model_path),
                                          model_type=stringValue(model_type),
                                          version_desc=stringValue(version_desc),
                                          current_stage=current_stage)
        request = metadata_service_pb2.RegisterModelVersionRequest(model_version=model_version)
        response = self.metadata_store_stub.registerModelVersion(request)
        return _unwrap_model_version_response(response)

    def delete_model_version_by_version(self, version, model_id) -> Status:
        """
        Delete registered model version by model version name .

        :param version: the model version name
        :param model_id: the model id corresponded to the model version
        :return: Status.OK if the model version is successfully deleted,
                 Status.ERROR if the model version does not exist otherwise.
        """
        request = metadata_service_pb2.ModelVersionNameRequest(name=version, model_id=model_id)
        response = self.metadata_store_stub.deleteModelVersionByVersion(request)
        return _unwrap_delete_response(response)

    def get_deployed_model_version(self, model_name) -> ModelVersionMeta:
        request = ModelNameRequest(name=model_name)
        response = self.metadata_store_stub.getDeployedModelVersion(request)
        return _unwrap_model_version_response(response)

    def get_latest_validated_model_version(self, model_name) -> ModelVersionMeta:
        request = ModelNameRequest(name=model_name)
        response = self.metadata_store_stub.getLatestValidatedModelVersion(request)
        return _unwrap_model_version_response(response)

    def get_latest_generated_model_version(self, model_name) -> ModelVersionMeta:
        request = ModelNameRequest(name=model_name)
        response = self.metadata_store_stub.getLatestGeneratedModelVersion(request)
        return _unwrap_model_version_response(response)

    '''project api'''

    def get_project_by_id(self, project_id) -> Optional[ProjectMeta]:
        """
        get a specific project in metadata store by project id

        :param project_id: the project id
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        request = metadata_service_pb2.IdRequest(id=project_id)
        response = self.metadata_store_stub.getProjectById(request)
        return _unwrap_project_response(response)

    def get_project_by_name(self, project_name) -> Optional[ProjectMeta]:
        """
        get a specific project in metadata store by project name
        :param project_name: the project name
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        request = metadata_service_pb2.NameRequest(name=project_name)
        response = self.metadata_store_stub.getProjectByName(request)
        return _unwrap_project_response(response)

    def register_project(self, name, uri: Text = None, properties: Properties = None) -> ProjectMeta:
        """
        register a project in metadata store.

        :param name: the name of the project
        :param uri: the uri of the project
        :param properties: the properties of the project
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object.
        """
        project_request = ProjectProto(name=name, uri=stringValue(uri), properties=properties)
        request = metadata_service_pb2.RegisterProjectRequest(project=project_request)
        response = self.metadata_store_stub.registerProject(request)
        return _unwrap_project_response(response)

    def update_project(self, project_name: Text, uri: Text = None, properties: Properties = None) -> Optional[ProjectMeta]:
        """
        update project in metadata store.

        :param project_name: the name of the project
        :param uri: the uri of the project
        :param properties: the properties of the project
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if update successfully.
        """
        request = metadata_service_pb2.UpdateProjectRequest(name=project_name, uri=stringValue(uri),
                                                            properties=properties)
        response = self.metadata_store_stub.updateProject(request)
        return _unwrap_project_response(response)

    def list_project(self, page_size, offset) -> Optional[List[ProjectMeta]]:
        """
        List registered projects in metadata store.

        :param page_size: the limitation of the listed projects.
        :param offset: the offset of listed projects.
        :return: List of :py:class:`ai_flow.meta.project_meta.ProjectMeta` objects,
                 return None if no projects to be listed.
        """
        request = metadata_service_pb2.ListRequest(page_size=page_size, offset=offset)
        response = self.metadata_store_stub.listProject(request)
        return _unwrap_project_list_response(response)

    def delete_project_by_id(self, project_id) -> Status:
        """
        Delete the registered project by project id .

        :param project_id: the project id
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        request = metadata_service_pb2.IdRequest(id=project_id)
        response = self.metadata_store_stub.deleteProjectById(request)
        return _unwrap_delete_response(response)

    def delete_project_by_name(self, project_name) -> Status:
        """
        Delete the registered project by project name .

        :param project_name: the project name
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        request = metadata_service_pb2.NameRequest(name=project_name)
        response = self.metadata_store_stub.deleteProjectByName(request)
        return _unwrap_delete_response(response)

    '''artifact api'''

    def get_artifact_by_id(self, artifact_id) -> Optional[ArtifactMeta]:
        """
        get a specific artifact in metadata store by artifact id.

        :param artifact_id: the artifact id
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """
        request = metadata_service_pb2.IdRequest(id=artifact_id)
        response = self.metadata_store_stub.getArtifactById(request)
        return _unwrap_artifact_response(response)

    def get_artifact_by_name(self, artifact_name) -> Optional[ArtifactMeta]:
        """
        get a specific artifact in metadata store by artifact name.

        :param artifact_name: the artifact name
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """
        request = metadata_service_pb2.NameRequest(name=artifact_name)
        response = self.metadata_store_stub.getArtifactByName(request)
        return _unwrap_artifact_response(response)

    def register_artifact(self, name: Text, artifact_type: Text = None, description: Text = None,
                          uri: Text = None, properties: Properties = None) -> ArtifactMeta:
        """
        register an artifact in metadata store.

        :param name: the name of the artifact
        :param artifact_type: the type of the artifact
        :param description: the description of the artifact
        :param uri: the uri of the artifact
        :param properties: the properties of the artifact
        :return: A single :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` object.
        """
        artifact_request = ArtifactProto(name=name, artifact_type=stringValue(artifact_type),
                                         description=stringValue(description), uri=stringValue(uri),
                                         properties=properties)
        request = metadata_service_pb2.RegisterArtifactRequest(artifact=artifact_request)
        response = self.metadata_store_stub.registerArtifact(request)
        return _unwrap_artifact_response(response)

    def update_artifact(self, artifact_name: Text, artifact_type: Text = None,
                        description: Text = None, uri: Text = None,
                        properties: Properties = None) -> Optional[ArtifactMeta]:
        """
        update artifact in metadata store.

        :param artifact_name: the name of the artifact
        :param artifact_type: the type of the artifact
        :param description: the description of the artifact
        :param uri: the batch uri of the artifact
        :param properties: the properties of the artifact
        :return: A single :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` object if update successfully.
        """
        request = metadata_service_pb2.UpdateArtifactRequest(name=artifact_name,
                                                             artifact_type=stringValue(artifact_type),
                                                             description=stringValue(description),
                                                             uri=stringValue(uri),
                                                             properties=properties)
        response = self.metadata_store_stub.updateArtifact(request)
        return _unwrap_artifact_response(response)

    def list_artifact(self, page_size, offset) -> Optional[List[ArtifactMeta]]:
        """
        List registered artifacts in metadata store.

        :param page_size: the limitation of the listed artifacts.
        :param offset: the offset of listed artifacts.
        :return: List of :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` objects,
                 return None if no artifacts to be listed.
        """
        request = metadata_service_pb2.ListRequest(page_size=page_size, offset=offset)
        response = self.metadata_store_stub.listArtifact(request)
        return _unwrap_artifact_list_response(response)

    def delete_artifact_by_id(self, artifact_id) -> Status:
        """
        Delete the registered artifact by artifact id .

        :param artifact_id: the artifact id
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """
        request = metadata_service_pb2.IdRequest(id=artifact_id)
        response = self.metadata_store_stub.deleteArtifactById(request)
        return _unwrap_delete_response(response)

    def delete_artifact_by_name(self, artifact_name) -> Status:
        """
        Delete the registered artifact by artifact name .

        :param artifact_name: the artifact name
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """
        request = metadata_service_pb2.NameRequest(name=artifact_name)
        response = self.metadata_store_stub.deleteArtifactByName(request)
        return _unwrap_delete_response(response)

    '''workflow api'''

    def register_workflow(self, name: Text, project_id: int, properties: Properties = None) -> WorkflowMeta:
        """
        Register a workflow in metadata store.

        :param name: the workflow name
        :param project_id: the id of project which contains the workflow
        :param properties: the workflow properties
        """
        workflow_request = WorkflowMetaProto(name=name,
                                             project_id=int64Value(project_id),
                                             properties=properties)
        request = metadata_service_pb2.RegisterWorkflowRequest(workflow=workflow_request)
        response = self.metadata_store_stub.registerWorkflow(request)
        return _unwrap_workflow_response(response)

    def get_workflow_by_name(self, project_name: Text, workflow_name: Text) -> Optional[WorkflowMeta]:
        """
        Get a workflow by specific project name and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """
        request = metadata_service_pb2.WorkflowNameRequest(workflow_name=workflow_name,
                                                           project_name=project_name)
        response = self.metadata_store_stub.getWorkflowByName(request)
        return _unwrap_workflow_response(response)

    def get_workflow_by_id(self, workflow_id: int) -> Optional[WorkflowMeta]:
        """
        Get a workflow by specific uuid

        :param workflow_id: the uuid of workflow
        """
        request = metadata_service_pb2.IdRequest(id=workflow_id)
        response = self.metadata_store_stub.getWorkflowById(request)
        return _unwrap_workflow_response(response)

    def list_workflows(self, project_name: Text, page_size: int, offset: int) -> Optional[List[WorkflowMeta]]:
        """
        List all workflows of the specific project

        :param project_name: the name of project which contains the workflow
        :param page_size      Limitation of listed workflows.
        :param offset        Offset of listed workflows.
        """
        request = metadata_service_pb2.ListWorkflowsRequest(project_name=project_name,
                                                            page_size=page_size,
                                                            offset=offset)
        response = self.metadata_store_stub.listWorkflows(request)
        return _unwrap_workflow_list_response(response)

    def delete_workflow_by_name(self, project_name: Text, workflow_name: Text) -> Status:
        """
        Delete the workflow by specific project and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """
        request = metadata_service_pb2.WorkflowNameRequest(workflow_name=workflow_name,
                                                           project_name=project_name)
        response = self.metadata_store_stub.deleteWorkflowByName(request)
        return _unwrap_delete_response(response)

    def delete_workflow_by_id(self, workflow_id: int) -> Status:
        """
        Delete the workflow by specific id

        :param workflow_id: the uuid of workflow
        """
        request = metadata_service_pb2.IdRequest(id=workflow_id)
        response = self.metadata_store_stub.deleteWorkflowById(request)
        return _unwrap_delete_response(response)

    def update_workflow(self, workflow_name: Text, project_name: Text,
                        properties: Properties = None) -> Optional[WorkflowMeta]:
        """
        Update the workflow

        :param workflow_name: the workflow name
        :param project_name: the name of project which contains the workflow
        :param properties: (Optional) the properties need to be updated
        """
        request = metadata_service_pb2.UpdateWorkflowRequest(workflow_name=workflow_name,
                                                             project_name=project_name,
                                                             properties=properties)
        response = self.metadata_store_stub.updateWorkflow(request)
        return _unwrap_workflow_response(response)
